/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TERATERM.EXE, keyboard routines */

#include "teraterm.h"
#include "tttypes.h"
#include "ttlib.h"
#include "ttsetup.h"
#include "ttcommon.h"
#include "ttwinman.h"
#include "ttime.h"

#include "keyboard.h"

int NewKeyId, NewKeyLen;
BYTE NewKeyStr[FuncKeyStrMax];

static char FuncKeyStr[IdF20-IdF6+1][FuncKeyStrMax];
static int FuncKeyLen[IdF20-IdF6+1];
BOOL OpenConv = FALSE;
BOOL AutoRepeatMode;
BOOL AppliKeyMode, AppliCursorMode;

  /*keyboard status*/
static int PreviousKey;

  /*key code map*/
static PKeyMap KeyMap = NULL;

  /* Character string sent by key pressiong */
static int CodeCount, CodeLength;
static char Code[101];
static BOOL CodeConv;

void SetKeyMap()
{
  char TempDir[MAXPATHLEN];
  char TempName[MAXPATHLEN];

  if ( strlen(ts.KeyCnfFN)==0 ) return;
  ExtractFileNameJ(ts.KeyCnfFN,TempName);
  ExtractDirNameJ(ts.KeyCnfFN,TempDir);

  if ( (stricmp(TempDir,ts.HomeDir)!=0) ||
     (stricmp(TempName,"KEYBOARD.CNF")!=0) )
  {
    if ( KeyMap==NULL )
      KeyMap = (PKeyMap)malloc(sizeof(TKeyMap));
    if ( KeyMap!=NULL )
    {
      if ( LoadTTSET() )
        (*ReadKeyboardCnf)(ts.KeyCnfFN, KeyMap);
      FreeTTSET();
    }
  }
  else if ( KeyMap!=NULL )
  {
    free(KeyMap);
    KeyMap = NULL;
  }
}

void ClearUserKey()
{
  int i;

  i = 0;
  while (i <= IdF20-IdF6)
  {
    FuncKeyLen[i] = 0;
    i++;
  }
}

void DefineUserKey()
{
  if ( (NewKeyLen==0) || (NewKeyLen>FuncKeyStrMax) ) return;

  if ((NewKeyId>=17) && (NewKeyId<=21))
    NewKeyId = NewKeyId-17;
  else if ((NewKeyId>=23) && (NewKeyId<=26))
    NewKeyId = NewKeyId-18;
  else if ((NewKeyId>=28) && (NewKeyId<=29))
    NewKeyId = NewKeyId-19;
  else if ((NewKeyId>=31) && (NewKeyId<=34))
    NewKeyId = NewKeyId-20;
  else
    return;

  memcpy(&FuncKeyStr[NewKeyId][0], NewKeyStr, NewKeyLen);
  FuncKeyLen[NewKeyId] = NewKeyLen;

}

void KeyDown(HWND HWin, WORD VKey, WORD Count, WORD Scan)
{
  WORD Key;
  MSG TM;
  BYTE KeyState[256];
  int i;

  /* for Win95J */
  if (OpenConv) return;

  if ( (VKey==VK_SHIFT) ||
       (VKey==VK_CONTROL) ||
       (VKey==VK_MENU) ) return;

  /* debug mode */
  if ( ts.Debug && (VKey == VK_ESCAPE) &&
       ShiftKey() )
  {
    MessageBeep(0);
    DebugFlag = ! DebugFlag;
    CodeCount = 0;
    PeekMessage((LPMSG)&TM,HWin,WM_CHAR,WM_CHAR,PM_REMOVE);
    return;
  }

  if ( ! AutoRepeatMode && (PreviousKey==VKey) )
  {
    PeekMessage((LPMSG)&TM,HWin,WM_CHAR,WM_CHAR,PM_REMOVE);
    return;
  }
  PreviousKey = VKey;

  if ( ShiftKey() ) Scan = Scan | 0x200;
  if ( ControlKey() ) Scan = Scan | 0x400;
  if ( AltKey() ) Scan = Scan | 0x800;

  CodeCount = Count;
  /* Windows keycode -> Tera Term keycode */
  Key = GetKeyCode(KeyMap,Scan);

  CodeConv = TRUE;
  if ( Key==0 )
  {
    switch (VKey) {
      case VK_BACK:
        CodeLength = 1;
        if ( ControlKey() )
        {
          if ( ts.BSKey==IdDEL )
            Code[0] = 0x08;
          else Code[0] = 0x7F;
        }
        else {
          if ( ts.BSKey==IdDEL )
            Code[0] = 0x7F;
          else Code[0] = 0x08;
        }
        break;
      case VK_RETURN: /* CR Key */
        CodeLength = 1;
        Code[0] = 0x0D;
        break;
      case VK_SPACE:
        if (ControlKey())
        {
          CodeLength = 1;
          Code[0] = 0;
        }
	else
          CodeLength = 0;
        break;
      default:
        CodeLength = 0;
    }
  }
  else
    if ( (IdF6<=Key) && (Key<=IdF20) && (FuncKeyLen[Key-IdF6]>0) )
    {
      memcpy(Code,&FuncKeyStr[Key-IdF6][0],FuncKeyLen[Key-IdF6]);
      CodeLength = FuncKeyLen[Key-IdF6];
      CodeConv = FALSE;
    }
    else
      GetKeyStr(HWin,KeyMap,Key,AppliKeyMode,AppliCursorMode,
                Code,&CodeLength,&CodeConv);

  if ( CodeLength==0 ) return;

  if ( VKey==VK_NUMLOCK )
  {
    /* keep NumLock LED status */
    GetKeyboardState((PBYTE)KeyState);
    KeyState[VK_NUMLOCK] = KeyState[VK_NUMLOCK] ^ 1;
    SetKeyboardState((PBYTE)KeyState);
  }

  PeekMessage((LPMSG)&TM,HWin,WM_CHAR,WM_CHAR,PM_REMOVE);

  if ( TalkStatus==IdTalkKeyb )
  {
    if ( CodeConv )
      for (i = 1 ; i <= CodeCount ; i++)
      {
        if ( ts.LocalEcho>0 )
          CommTextEcho(&cv,Code,CodeLength);
        CommTextOut(&cv,Code,CodeLength);
      }
    else
      for (i = 1 ; i <= CodeCount ; i++)
      {
        CommBinaryOut(&cv,Code,CodeLength);
        if ( ts.LocalEcho>0 )
          CommBinaryEcho(&cv,Code,CodeLength);
      }
  }
}

void KeyUp(WORD VKey)
{
  if ( PreviousKey == VKey ) PreviousKey = 0;
}

BOOL ShiftKey()
{
  return ((GetAsyncKeyState(VK_SHIFT) & 0xFFFFFF80) != 0);
}

BOOL ControlKey()
{
  return ((GetAsyncKeyState(VK_CONTROL) & 0xFFFFFF80) != 0);
}

BOOL AltKey()
{
  return ((GetAsyncKeyState(VK_MENU) & 0xFFFFFF80) != 0);
}

void InitKeyboard()
{
  KeyMap = NULL;
  ClearUserKey();
  OpenConv = FALSE;
  PreviousKey = 0;
}

void EndKeyboard()
{
  if (KeyMap != NULL)
    free(KeyMap);
}


