/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TERATERM.EXE, Scroll bars of VT window */
#include "teraterm.h"
#include "tttypes.h"
#include "ttwinman.h"
#include "buffer.h"
#include "vtdisp.h"
#include <stdio.h>

int WinOrgX, WinOrgY, NewOrgX, NewOrgY;

int ScrollCount = 0;
int dScroll = 0;
int SRegionTop;
int SRegionBottom;

void AdjustScrollBar() /* called by ChangeWindowSize() */
{
  LONG XRange, YRange;
  int ScrollPosX, ScrollPosY;

  if (NumOfColumns-WinWidth>0)
    XRange = NumOfColumns-WinWidth;
  else
    XRange = 0;

  if (BuffEnd-WinHeight>0)
    YRange = BuffEnd-WinHeight;
  else
    YRange = 0;

  ScrollPosX = GetScrollPos(HVTWin,SB_HORZ);
  ScrollPosY = GetScrollPos(HVTWin,SB_VERT);
  if (ScrollPosX > XRange)
    ScrollPosX = XRange;
  if (ScrollPosY > YRange)
    ScrollPosY = YRange;

  WinOrgX = ScrollPosX;
  WinOrgY = ScrollPosY-PageStart;
  NewOrgX = WinOrgX;
  NewOrgY = WinOrgY;

  DontChangeSize = TRUE;

  SetScrollRange(HVTWin,SB_HORZ,0,XRange,FALSE);

  if ((YRange == 0) && (ts.EnableScrollBuff>0))
  {
    SetScrollRange(HVTWin,SB_VERT,0,1,FALSE);
  }
  else {
    SetScrollRange(HVTWin,SB_VERT,0,YRange,FALSE);
  }

  SetScrollPos(HVTWin,SB_HORZ,ScrollPosX,TRUE);
  SetScrollPos(HVTWin,SB_VERT,ScrollPosY,TRUE);

  DontChangeSize = FALSE;  
}

void ScrollToCursor(int CurX, int CurY)
{
  if (CurX < NewOrgX)
    NewOrgX = CurX;
  else if (CurX >= NewOrgX+WinWidth)
    NewOrgX = CurX + 1 - WinWidth;

  if (CurY < NewOrgY)
    NewOrgY = CurY;
  else if (CurY >= NewOrgY+WinHeight)
    NewOrgY = CurY + 1 - WinHeight;
}

void Scroll1Line()
{
  int i;
  LONG SrcPtr, DestPtr;

  UpdateStr();

  if ((CursorTop == 0) && (CursorBottom == NumOfLines-1))
  {
    WinOrgY--;
    BuffScroll(1);
    CountScroll();
  }
  else if ((CursorTop<=CursorY) && (CursorY<=CursorBottom))
  {
    DestPtr = GetLinePtr(PageStart+CursorTop);
    for (i = CursorTop+1 ; i<=CursorBottom ; i++)
    {
      SrcPtr = NextLinePtr(DestPtr);
      memmove(&(CodeBuff[DestPtr]),&(CodeBuff[SrcPtr]),NumOfColumns);
      memmove(&(AttrBuff[DestPtr]),&(AttrBuff[SrcPtr]),NumOfColumns);
      memmove(&(AttrBuff2[DestPtr]),&(AttrBuff2[SrcPtr]),NumOfColumns);
      DestPtr = SrcPtr;
    }
    memset(&(CodeBuff[SrcPtr]),0x20,NumOfColumns);
    memset(&(AttrBuff[SrcPtr]),AttrDefault,NumOfColumns);
    memset(&(AttrBuff2[SrcPtr]),AttrDefault2,NumOfColumns);

    if ((dScroll<0) ||
        (dScroll>0) &&
        ((SRegionTop!=CursorTop) ||
	 (SRegionBottom!=CursorBottom)))
      UpdateScroll();
    SRegionTop = CursorTop;
    SRegionBottom = CursorBottom;
    dScroll++;
    CountScroll();
  }
}

void CountScroll()
{
  ScrollCount++;
  if (ScrollCount>=ts.ScrollThreshold) UpdateScroll();
}

void UpdateScroll()
{
  int d;
  RECT R;

  ScrollCount = 0;

  /* Update partial scroll */
  if (dScroll != 0)
  {
    d = dScroll * FontHeight;
    R.left = 0;
    R.right = ScreenWidth;
    R.top = (SRegionTop-WinOrgY)*FontHeight;
    R.bottom = (SRegionBottom+1-WinOrgY)*FontHeight;
    ScrollWindow(HVTWin,0,-d,&R,&R);
    dScroll = 0;
  }

  /* Update normal scroll */
  if (NewOrgX < 0) NewOrgX = 0;
  if (NewOrgX>NumOfColumns-WinWidth)
    NewOrgX = NumOfColumns-WinWidth;
  if (NewOrgY < -PageStart) NewOrgY = -PageStart;
  if (NewOrgY>BuffEnd-WinHeight-PageStart)
    NewOrgY = BuffEnd-WinHeight-PageStart;

  if ((NewOrgX==WinOrgX) &&
      (NewOrgY==WinOrgY)) return;

  if (NewOrgX==WinOrgX)
  {
    d = (NewOrgY-WinOrgY) * FontHeight;
    ScrollWindow(HVTWin,0,-d,NULL,NULL);
  }
  else if (NewOrgY==WinOrgY)
  {
    d = (NewOrgX-WinOrgX) * FontWidth;
    ScrollWindow(HVTWin,-d,0,NULL,NULL);
  }
  else
    InvalidateRect(HVTWin,NULL,TRUE);

  /* Update scroll bars */
  if (NewOrgX!=WinOrgX)
    SetScrollPos(HVTWin,SB_HORZ,NewOrgX,TRUE);

  if (NewOrgY!=WinOrgY)
  {
    if ((BuffEnd==WinHeight) &&
        (ts.EnableScrollBuff>0))
      SetScrollRange(HVTWin,SB_VERT,0,1,TRUE);
    else
      SetScrollRange(HVTWin,SB_VERT,0,BuffEnd-WinHeight,FALSE);
    SetScrollPos(HVTWin,SB_VERT,NewOrgY+PageStart,TRUE);
  }

  WinOrgX = NewOrgX;
  WinOrgY = NewOrgY;
}

void ScrollUp1Line()
{
  int i;
  LONG SrcPtr, DestPtr;

  if ((CursorTop<=CursorY) && (CursorY<=CursorBottom))
  {
    UpdateStr();
    if ((dScroll > 0) ||
        (dScroll < 0) &&
        ((SRegionTop != CursorTop) ||
         (SRegionBottom != CursorBottom)))
      UpdateScroll();

    DestPtr = GetLinePtr(PageStart+CursorBottom);
    for (i = CursorBottom-1 ; i >= CursorTop ; i--)
    {
      SrcPtr = PrevLinePtr(DestPtr);
      memcpy(&(CodeBuff[DestPtr]),&(CodeBuff[SrcPtr]),NumOfColumns);
      memcpy(&(AttrBuff[DestPtr]),&(AttrBuff[SrcPtr]),NumOfColumns);
      memcpy(&(AttrBuff2[DestPtr]),&(AttrBuff2[SrcPtr]),NumOfColumns);
      DestPtr = SrcPtr;
    }
    memset(&(CodeBuff[SrcPtr]),0x20,NumOfColumns);
    memset(&(AttrBuff[SrcPtr]),AttrDefault,NumOfColumns);
    memset(&(AttrBuff2[SrcPtr]),AttrDefault2,NumOfColumns);

    SRegionTop = CursorTop;
    SRegionBottom = CursorBottom;
    dScroll--;
    CountScroll();
  }
}

void ScrollHomePos()
{
  NewOrgX = 0;
  NewOrgY = 0;
  UpdateScroll();
}

void AutoScroll(POINT p)
{
  int X, Y;

  X = (p.x + FontWidth / 2) / FontWidth;
  Y = p.y / FontHeight;
  if (X<0)
    NewOrgX = WinOrgX + X;
  else if (X>=WinWidth)
    NewOrgX = NewOrgX + X - WinWidth + 1;
  if (Y<0)
    NewOrgY = WinOrgY + Y;
  else if (Y>=WinHeight)
    NewOrgY = NewOrgY + Y - WinHeight + 1;

  UpdateScroll();
}
