/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TTTEK.DLL, TEK escape sequences */
#include "teraterm.h"
#include "tttypes.h"
#include "tektypes.h"
#include "ttcommon.h"

#include "tekesc.h"

void Log1Byte(PComVar cv, BYTE b)
{
  ((PCHAR)(cv->LogBuf))[cv->LogPtr] = b;
  cv->LogPtr++;
  if (cv->LogPtr >= InBuffSize)
    cv->LogPtr = cv->LogPtr - InBuffSize;
  if (cv->LCount >= InBuffSize)
  {
    cv->LCount = InBuffSize;
    cv->LStart = cv->LogPtr;
  }
  else cv->LCount++;
}

void ChangeTextSize(PTEKVar tk, PTTSet ts)
{
  int i;

  SelectObject(tk->MemDC,tk->TEKFont[tk->TextSize]);
  tk->FontWidth = tk->FW[tk->TextSize];
  tk->FontHeight = tk->FH[tk->TextSize];

  for (i = 0; i <= 1; i++)
    tk->Dx[i] = tk->FontWidth;

  if (tk->Active) ChangeCaret(tk, ts);
}

void BackSpace(PTEKVar tk, PComVar cv)
{
  tk->CaretX = tk->CaretX - tk->FontWidth;
  if (tk->CaretX < 0) tk->CaretX = 0;
  if (cv->HLogBuf != NULL) Log1Byte(cv,BS);
}

void LineFeed(PTEKVar tk, PComVar cv)
{
  tk->CaretY = tk->CaretY + tk->FontHeight;
  if (tk->CaretY > tk->ScreenHeight - tk->FontHeight)
  {
    tk->CaretY = 0;
    if ((tk->CaretOffset==0) || (tk->CaretX < tk->CaretOffset))
    {
      tk->CaretOffset = tk->ScreenWidth >> 1;
      tk->CaretX = tk->CaretX + tk->CaretOffset;
      if (tk->CaretX >= tk->ScreenWidth)
	tk->CaretX = tk->CaretOffset;
    }
    else {
      tk->CaretX = tk->CaretX - tk->CaretOffset;
      tk->CaretOffset = 0;
    }
  }
  if (cv->HLogBuf != NULL) Log1Byte(cv,LF);
}

void CarriageReturn(PTEKVar tk, PComVar cv)
{
  tk->CaretX = tk->CaretOffset;
  if (cv->HLogBuf != NULL) Log1Byte(cv,CR);
}

void Tab(PTEKVar tk, PComVar cv)
{
  if (cv->HLogBuf != NULL) Log1Byte(cv,HT);
  tk->CaretX = tk->CaretX + (tk->FontWidth << 3);
  if (tk->CaretX >= tk->ScreenWidth)
  {
    CarriageReturn(tk,cv);
    LineFeed(tk,cv);
  }
}

void EnterVectorMode(PTEKVar tk)
{
  tk->MoveFlag = TRUE;
  if (tk->DispMode == IdAlphaMode)
    tk->CaretY = tk->CaretY + tk->FontHeight - 1;
  tk->Drawing = FALSE;
  tk->DispMode = IdVectorMode;
}

void EnterPlotMode(PTEKVar tk)
{
  if (tk->DispMode == IdAlphaMode)
  {
    tk->CaretY = tk->CaretY + tk->FontHeight - 1;
    tk->CaretOffset = 0;
  }
  tk->PlotX =
    (int)((float)(tk->CaretX) / (float)(tk->ScreenWidth) *
          (float)(ViewSizeX));
  tk->PlotY =
    (int)((float)(tk->ScreenHeight-1-tk->CaretY) /
	  (float)(tk->ScreenHeight) * (float)(ViewSizeY));
  tk->DispMode = IdPlotMode;
  tk->JustAfterRS = TRUE;
}

void EnterAlphaMode(PTEKVar tk)
{
  if ((tk->DispMode == IdVectorMode) ||
      (tk->DispMode == IdPlotMode))
  {
    tk->CaretY = tk->CaretY - tk->FontHeight + 1;
    tk->CaretOffset = 0;
  }
  tk->DispMode = IdAlphaMode;
  tk->Drawing = FALSE;
}

void EnterUnknownMode(PTEKVar tk)
{
  if ((tk->DispMode == IdVectorMode) ||
      (tk->DispMode == IdPlotMode))
  {
    tk->CaretY = tk->CaretY - tk->FontHeight + 1;
    tk->CaretOffset = 0;
  }
  tk->DispMode = IdUnknownMode;
}

void TEKMoveTo(PTEKVar tk, PTTSet ts, int X, int Y, BOOL Draw)
{
  HDC DC;
  HPEN OldPen;

  if (Draw)
  {
    DC = GetDC(tk->HWin);
    OldPen = SelectObject(DC,tk->Pen);
    SetBkColor(DC,ts->TEKColor[1]);
    SetBkMode(DC, 1);
    if ((tk->CaretX == X) && (tk->CaretY==Y))
    {
      SetPixel(DC,X,Y,tk->PenColor);
      SetPixel(tk->MemDC,X,Y,tk->MemPenColor);
    }
    else {
      MoveToEx(DC,tk->CaretX,tk->CaretY,NULL);
      LineTo(DC,X,Y);
      MoveToEx(tk->MemDC,tk->CaretX,tk->CaretY,NUL);
      LineTo(tk->MemDC,X,Y);
    }
    SelectObject(DC,OldPen);
    ReleaseDC(tk->HWin,DC);
  }
  tk->CaretX = X;
  tk->CaretY = Y;
}

void Draw(PTEKVar tk, PComVar cv, PTTSet ts, BYTE b)
{
  int X, Y;

  if (! tk->Drawing)
  {
    tk->LoXReceive = FALSE;
    tk->LoCount = 0;
    tk->Drawing = TRUE;
  }

  if (b<=0x1f)
  {
    CommInsert1Byte(cv,b);
    return;
  }
  else if ((b>=0x20) && (b<=0x3f))
  {
    if (tk->LoCount==0)
      tk->HiY = b - 0x20;
    else tk->HiX = b - 0x20;
  }
  else if ((b>=0x40) && (b<=0x5f))
  {
    tk->LoX = b - 0x40;
    tk->LoXReceive = TRUE;
  }
  else if ((b>=0x60) && (b<=0x7f))
  {
    tk->LoA = tk->LoB;
    tk->LoB = b - 0x60;
    tk->LoCount++;
  }

  if (! tk->LoXReceive) return;

  tk->Drawing = FALSE;
  tk->ParseMode = ModeFirst;
  if (tk->LoCount > 0) tk->LoY = tk->LoB;
  if (tk->LoCount > 1) tk->Extra = tk->LoA;

  X = (int)((float)(tk->HiX*128 + tk->LoX*4 + (tk->Extra & 3)) /
	    (float)(ViewSizeX) * (float)(tk->ScreenWidth));
  Y = tk->ScreenHeight - 1 -
    (int)((float)(tk->HiY * 128 + tk->LoY * 4 + tk->Extra / 4) /
          (float)(ViewSizeY) * (float)(tk->ScreenHeight));
  TEKMoveTo(tk,ts,X,Y,! tk->MoveFlag);
  tk->MoveFlag = FALSE;
}

void Plot(PTEKVar tk, PTTSet ts, BYTE b)
{
  int X, Y;

  X = 0;
  Y = 0;
  switch (b) {
    case 'A': X = 4; break;
    case 'B': X = -4; break;
    case 'D': Y = 4; break;
    case 'E':
      X = 4;
      Y = 4;
      break;
    case 'F':
      X = -4;
      Y = 4;
      break;
    case 'H': Y = -4; break;
    case 'I':
      X = 4;
      Y = -4;
      break;
    case 'J':
      X = -4;
      Y = -4;
      break;
    default:
      return;
  }
  tk->PlotX = tk->PlotX + X;
  if (tk->PlotX < 0) tk->PlotX = 0;
  if (tk->PlotX >= ViewSizeX)
    tk->PlotX = ViewSizeX - 1;
  tk->PlotY = tk->PlotY + Y;
  if (tk->PlotY < 0) tk->PlotY = 0;
  if (tk->PlotY >= ViewSizeY)
    tk->PlotY = ViewSizeY - 1;
  X = (int)((float)(tk->PlotX) / (float)(ViewSizeX) *
            (float)(tk->ScreenWidth));
  Y = tk->ScreenHeight - 1 -
    (int)((float)(tk->PlotY) / (float)(ViewSizeY) *
	  (float)(tk->ScreenHeight));
  TEKMoveTo(tk,ts,X,Y,tk->PenDown);
}

void DispChar(PTEKVar tk, PComVar cv, BYTE b)
{
  tk->R.left = tk->CaretX;
  tk->R.right = tk->CaretX + tk->FontWidth;
  tk->R.top = tk->CaretY;
  tk->R.bottom = tk->CaretY + tk->FontHeight;
  ExtTextOut(tk->MemDC,tk->CaretX,tk->CaretY,0,&(tk->R),&b,1,&(tk->Dx[0]));
  InvalidateRect(tk->HWin,&(tk->R),FALSE);
  tk->CaretX = tk->R.right;

  if (cv->HLogBuf != NULL) Log1Byte(cv,b);

  if (tk->CaretX > tk->ScreenWidth - tk->FontWidth)
  {
    CarriageReturn(tk,cv);
    LineFeed(tk,cv);
  }
}

void ParseFirst(PTEKVar tk, PTTSet ts, PComVar cv, BYTE b)
{
  if (tk->DispMode == IdAlphaMode)
    switch (b) {
      case 0x07: MessageBeep(0); break;
      case 0x08: BackSpace(tk,cv); break;
      case 0x09: Tab(tk,cv); break;
      case 0x0a: LineFeed(tk,cv); break;
      case 0x0b:
      case 0x0c: break;
      case 0x0d:
        CarriageReturn(tk,cv);
        if (ts->CRReceive==IdCRLF)
          CommInsert1Byte(cv,0x0A);
        break;
      case 0x18:
	if (ts->AutoWinSwitch > 0)
	  tk->ChangeEmu = IdVT; /* Enter VT Mode */
	break;
      case 0x19:
      case 0x1a: break;
      case 0x1b:
        tk->ParseMode = ModeEscape;
        return;
      case 0x1c: EnterUnknownMode(tk); break;
      case 0x1d: EnterVectorMode(tk); break;
      case 0x1e: EnterPlotMode(tk); break;
      case 0x1f: break;
      default:
        if (b<=0x06) break;
        if ((b>=0x0e) && (b<=0x17)) break;
        if (tk->Drawing) Draw(tk,cv,ts,b);
          else DispChar(tk,cv,b);
    }
  else if (tk->DispMode==IdVectorMode)
    switch (b) {
      case 0x07:
	if (tk->MoveFlag) tk->MoveFlag = FALSE;
          else MessageBeep(0);
	break;
      case 0x0d:
        CommInsert1Byte(cv,0x0d);
        EnterAlphaMode(tk); /* EnterAlphaMode */
        break;
      case 0x18:
	if (ts->AutoWinSwitch > 0)
	  tk->ChangeEmu = IdVT; /* Enter VT Mode */
	break;
      case 0x19:
      case 0x1a: break;
      case 0x1b:
        tk->ParseMode = ModeEscape;
        return;
      case 0x1c: EnterUnknownMode(tk); break;
      case 0x1d: EnterVectorMode(tk); break;
      case 0x1e: EnterPlotMode(tk); break;
      case 0x1f: EnterAlphaMode(tk); break; /* EnterAlphaMode */
      default:
	if (b<=0x06) break;
	if ((b>=0x08) && (b<=0x0c)) break;
	if ((b>=0x0e) && (b<=0x17)) break;
        Draw(tk,cv,ts,b);
    }
  else if (tk->DispMode==IdPlotMode)
    switch (b) {
      case 0x07: MessageBeep(0); break;
      case 0x0d:
        CommInsert1Byte(cv,0x0d);
        EnterAlphaMode(tk); /* EnterAlphaMode */
	break;
      case 0x18:
	if (ts->AutoWinSwitch > 0)
	  tk->ChangeEmu = IdVT; /* Enter VT Mode */
	break;
      case 0x19:
      case 0x1a: break;
      case 0x1b:
        tk->ParseMode = ModeEscape;
        return;
      case 0x1c: EnterUnknownMode(tk); break;
      case 0x1d: EnterVectorMode(tk); break;
      case 0x1e: EnterPlotMode(tk); break;
      case 0x1f: EnterAlphaMode(tk); break; /* EnterAlphaMode */
      default:
	if (b<=0x06) break;
        if ((b>=0x08) && (b<=0x0c)) break;
	if ((b>=0x0e) && (b<=0x17)) break;

        if (tk->JustAfterRS)
        {
          if (b==0x20)
            tk->PenDown = FALSE;
          else if (b=='P')
            tk->PenDown = TRUE;
          else
            EnterUnknownMode(tk);
          tk->JustAfterRS = FALSE;
        }
        else             
          Plot(tk,ts,b);
    }
  else
    switch (b) {
      case 0x1f: EnterAlphaMode(tk); break;
    }

  tk->ParseMode = ModeFirst;
}

void SelectCode(PTEKVar tk, PTTSet ts, BYTE b)
{
  switch (b) {
    default:
      if (b <= US) return;
    case ' ':
      if (tk->SelectCodeFlag == '#')
	tk->SelectCodeFlag = b;
      return;
    case '!':
      if (tk->SelectCodeFlag == '%')
	tk->SelectCodeFlag = b;
      return;
    case '1':
    case '2':
      if ((tk->SelectCodeFlag == ' ') ||
          (tk->SelectCodeFlag == '!'))
	tk->ChangeEmu = IdVT; /* enter VT mode */
      break;
    case '3':
      if (tk->SelectCodeFlag == '!')
	tk->ChangeEmu = IdVT; /* enter VT mode */
      break;
    case '5':
      if (tk->SelectCodeFlag == ' ')
	tk->ChangeEmu = IdVT; /* enter VT mode */
      break;
  }
  if ((0x20<=b) && (b<0x7f)) tk->ParseMode = ModeFirst;
  tk->SelectCodeFlag = 0;
  if (ts->AutoWinSwitch > 0) tk->ChangeEmu = 0;
}

COLORREF ColorIndex(COLORREF Fore, COLORREF Back, WORD w)
{
  switch (w) {
    case 1: return Fore;
    case 2: return RGB(255,  0,  0); /* Red */
    case 3: return RGB(  0,255,  0); /* Green */
    case 4: return RGB(  0,  0,255); /* Blue */
    case 5: return RGB(  0,255,255); /* Cyan */
    case 6: return RGB(255,  0,255); /* Magenta */
    case 7: return RGB(255,255,  0); /* Yellow */
    case 8: return RGB(255,128,  0); /* Orange */
    case 9: return RGB(128,255,  0); /* Green-Yellow */
   case 10: return RGB(  0,255,128); /* Green-Cyan */
   case 11: return RGB(  0,128,255); /* Blue-Cyan */
   case 12: return RGB(128,  0,255); /* Blue-Magenta */
   case 13: return RGB(255,  0,128); /* Red-Magenta */
   case 14: return RGB( 85, 85, 85); /* Dark gray */
   case 15: return RGB(170,170,170); /* Light gray */
  }
  return Back;
}

void SetLineIndex(PTEKVar tk, PTTSet ts, WORD w)
{
  HWND TempDC;

  tk->PenColor = ColorIndex(ts->TEKColor[0],ts->TEKColor[1],w);
  TempDC = GetDC(tk->HWin);
  tk->PenColor = GetNearestColor(TempDC,tk->PenColor);
  ReleaseDC(tk->HWin,TempDC);
  if (ts->TEKColorEmu > 0)
    tk->MemPenColor = tk->PenColor;
  else
    if (tk->PenColor == ts->TEKColor[1])
      tk->MemPenColor = tk->MemBackColor;
    else
      tk->MemPenColor = tk->MemForeColor;

  DeleteObject(tk->Pen);
  tk->Pen = CreatePen(tk->ps,1,tk->PenColor);
  
  SelectObject(tk->MemDC, tk->OldMemPen);
  DeleteObject(tk->MemPen);
  tk->MemPen = CreatePen(tk->ps,1,tk->MemPenColor);
  tk->OldMemPen = SelectObject(tk->MemDC,tk->MemPen);
}

void SetTextIndex(PTEKVar tk, PTTSet ts, WORD w)
{
  HWND TempDC;

  tk->TextColor = ColorIndex(ts->TEKColor[0],ts->TEKColor[1],w);
  TempDC = GetDC(tk->HWin);
  tk->TextColor = GetNearestColor(TempDC,ts->TEKColor[0]);
  ReleaseDC(tk->HWin,TempDC);
  if (ts->TEKColorEmu > 0)
    tk->MemTextColor = ts->TEKColor[0];
  else
    if (tk->TextColor != ts->TEKColor[1])
      tk->MemTextColor = tk->MemBackColor;
    else
      tk->MemTextColor = tk->MemForeColor;

  SetTextColor(tk->MemDC, tk->MemTextColor);
}

void SetLineStyle(PTEKVar tk, BYTE b)
{
  switch (b) {
    case 0: tk->ps = PS_SOLID; break;
    case 1: tk->ps = PS_DOT; break;
    case 2: tk->ps = PS_DASHDOT; break;
    case 3: tk->ps = PS_DASH; break;
    case 4: tk->ps = PS_DASH; break;
    case 5: tk->ps = PS_DASHDOTDOT; break;
    case 6: tk->ps = PS_DASHDOT; break;
    case 7: tk->ps = PS_DASH; break;
    case 8: tk->ps = PS_DASHDOTDOT; break;
    default:
      tk->ps = PS_SOLID;
  }
  if (tk->Pen != NULL) DeleteObject(tk->Pen);
  tk->Pen = CreatePen(tk->ps,1,tk->PenColor);

  SelectObject(tk->MemDC, tk->OldMemPen);
  if (tk->MemPen != NULL) DeleteObject(tk->MemPen);
  tk->MemPen = CreatePen(tk->ps,1,tk->MemPenColor);
  tk->OldMemPen = SelectObject(tk->MemDC,tk->MemPen);
}

void TwoOpCode(PTEKVar tk, PTTSet ts, PComVar cv, BYTE b)
{
  if (tk->OpCount == 2)
  {
    tk->OpCount--;
    tk->Op2OC = b << 8;
    return;
  }
  else if (tk->OpCount == 1)
  {
    if (b <= 0x1a) return;
    else if ((b>=0x1b) && (b<=0x1f))
    {
      CommInsert1Byte(cv,b);
      tk->ParseMode = ModeFirst;
      return;
    }
    else {
      tk->OpCount--;
      tk->Op2OC = tk->Op2OC + b;
      switch (tk->Op2OC) {
        case IdMove: tk->PrmCountMax = 0; break;
        case IdDraw: tk->PrmCountMax = 0; break;
        case IdSetLineIndex: tk->PrmCountMax = 1; break;
        case IdSetTextIndex: tk->PrmCountMax = 1; break;
        case IdSetLineStyle: tk->PrmCountMax = 1; break;
        default:
          tk->PrmCountMax = 0;
      }
      tk->Prm2OC = 0;
      tk->PrmCount = 0;
      if (tk->PrmCountMax > 0) return;
    }
  }

  if (tk->PrmCount < tk->PrmCountMax)
  {
    if (b <= 0x1a) return;
    else if ((b>=0x1b) && (b<=0x1f))
    {
      CommInsert1Byte(cv,b);
      tk->PrmCount = tk->PrmCountMax;
      tk->ParseMode = ModeFirst;
    }
    else if ((b>=0x20) && (b<=0x2f))
    {  /* LoI (minus) */
      tk->Prm2OC = - (tk->Prm2OC*16 + b - 0x20);
      tk->PrmCount++;
    }
    else if ((b>=0x30) && (b<=0x3f))
    {  /* LoI (plus) */
      tk->Prm2OC = tk->Prm2OC*16 + b - 0x30;
      tk->PrmCount++;
    }
    else if ((b>=0x40) && (b<=0x7f))
      tk->Prm2OC = tk->Prm2OC*64 + b - 0x40;
  }

  if (tk->PrmCount < tk->PrmCountMax) return;

  switch (tk->Op2OC) {
    case IdMove:
      tk->LoXReceive = FALSE;
      tk->LoCount = 0;
      tk->Drawing = TRUE;
      tk->MoveFlag = TRUE;
      break;
    case IdDraw:
      tk->LoXReceive = FALSE;
      tk->LoCount = 0;
      tk->Drawing = TRUE;
      tk->MoveFlag = FALSE;
      break;
    case IdSetLineIndex:
      SetLineIndex(tk,ts,tk->Prm2OC);
      break;
    case IdSetTextIndex:
      SetTextIndex(tk,ts,tk->Prm2OC);
      break;
    case IdSetLineStyle:
      SetLineStyle(tk,(BYTE)(tk->Prm2OC));
      break;
  }

  tk->ParseMode = ModeFirst;
}

  void EscPage(PTEKVar tk) /* Clear Screen and enter ALPHAMODE */
  {
    SetLineStyle(tk,0);
    GetClientRect(tk->HWin,&(tk->R));
    FillRect(tk->MemDC,&(tk->R),tk->MemBackGround);
    InvalidateRect(tk->HWin,&(tk->R),FALSE);
    EnterAlphaMode(tk);
    tk->CaretX = 0;
    tk->CaretOffset = 0;
    tk->CaretY = 0;
  }

void TEKEscape(PTEKVar tk, PTTSet ts, PComVar cv, BYTE b)
{  
  switch (b) {
    case 0x00: return;
    case 0x0d: return;
    case 0x0a: return;
    case 0x0C: EscPage(tk); break;
    case 0x1A:
      tk->GIN = TRUE;
      /* Capture mouse */
      SetCapture(tk->HWin);
      EnterAlphaMode(tk);
      break;
    case 0x1B: return;
    case 0x1C:
    case 0x1D:
    case 0x1E:
    case 0x1F:
      CommInsert1Byte(cv,b); break;
    case '#':
      tk->ParseMode = ModeSelectCode;
      tk->SelectCodeFlag = b;
      return;
    case '%':
      tk->ParseMode = ModeSelectCode;
      tk->SelectCodeFlag = b;
      return;
    case '2':
      if (ts->AutoWinSwitch > 0)
	tk->ChangeEmu = IdVT; /* Enter VT Mode */
      break;
    default:
      if ((b>='8') && (b<=';'))
      {
        if (tk->TextSize != b-0x38)
        {
          tk->TextSize = b-0x38;
          ChangeTextSize(tk,ts);
        }
	break;
      }
      else if ((b>='I') && (b<='Z'))
      {
        tk->ParseMode = Mode2OC;
        tk->OpCount = 2;
        TwoOpCode(tk,ts,cv,b);
        return;
      }
      else if ((b>=0x60) && (b<=0x6f))
      {
	SetLineStyle(tk,(b-0x60) & 7);
	break;
      }
    case '[':
      if (ts->AutoWinSwitch > 0)
      {
        CommInsert1Byte(cv,0x5B);
        CommInsert1Byte(cv,0x1B);
        tk->ChangeEmu = IdVT;
      }
      break;
    case 0x7f: return;
  }
  tk->ParseMode = ModeFirst;
}

