/* Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved. */

/* TERATERM.EXE, VT window */

#include "teraterm.h"
#include "tttypes.h"
#include <winsock.h>
#include "ttcommon.h"
#include "ttwinman.h"
#include "ttsetup.h"
#include "keyboard.h"
#include "buffer.h"
#include "vtterm.h"
#include "vtdisp.h"
#include "scroll.h"
#include "ttdialog.h"
#include "ttime.h"
#include "commlib.h"
#include "clipboar.h"
#include "ttftypes.h"
#include "filesys.h"
#include "telnet.h"
#include "tektypes.h"
#include "tek.h"
#include "teraprn.h"
#include "ttdde.h"
#include "commdlg.h"

#include "vtwin.h"

#define Id_Menu 100
#define IdVTIcon 200
#define Id_Accs 100

/* mouse buttons */
#define IdLeftButton 0
#define IdMiddleButton 1
#define IdRightButton 2

BEGIN_MESSAGE_MAP( CVTWindow, CFrameWnd )
  //{{AFX_MSG_MAP( CVTWindow )
  ON_WM_ACTIVATE()
  ON_WM_CHAR()
  ON_WM_CLOSE()
  ON_WM_DESTROY()
  ON_WM_DROPFILES()
  ON_WM_GETMINMAXINFO()
  ON_WM_HSCROLL()
  ON_WM_INITMENUPOPUP()
  ON_WM_KEYDOWN()
  ON_WM_KEYUP()
  ON_WM_LBUTTONDBLCLK()
  ON_WM_LBUTTONDOWN()
  ON_WM_LBUTTONUP()
  ON_WM_MBUTTONDOWN()
  ON_WM_MBUTTONUP()
  ON_WM_MOUSEMOVE()
  ON_WM_MOVE()
  ON_WM_NCLBUTTONDBLCLK()
  ON_WM_NCRBUTTONDOWN()
  ON_WM_PAINT()
  ON_WM_RBUTTONDOWN()
  ON_WM_RBUTTONUP()
  ON_WM_SIZE()
  ON_WM_SYSCHAR()
  ON_WM_SYSCOMMAND()
  ON_WM_SYSKEYDOWN()
  ON_WM_SYSKEYUP()
  ON_WM_TIMER()
  ON_WM_VSCROLL()

  ON_MESSAGE(WM_IME_REPORT,OnIMEReport)
  ON_MESSAGE(WM_USER+MsgAccelCommand,OnAccelCommand)
  ON_MESSAGE(WM_USER+MsgChangeMenu,OnChangeMenu)
  ON_MESSAGE(WM_USER+MsgCloseIME,OnCloseIME)
  ON_MESSAGE(WM_USER+MsgComNotify,OnComNotify)
  ON_MESSAGE(WM_USER+MsgComOpen,OnComOpen)
  ON_MESSAGE(WM_USER+MsgComStart,OnComStart)
  ON_MESSAGE(WM_USER+MsgDlgHelp2,OnDlgHelp)
  ON_MESSAGE(WM_USER+MsgDdeEnd,OnDdeEnd)

  ON_COMMAND(CM_FILENEWCONNECTION,CMFileNewConnection)
  ON_COMMAND(CM_FILEDISCONNECT,CMFileDisconnect)
  ON_COMMAND(CM_FILELOG,CMFileLog)
  ON_COMMAND(CM_FILESEND,CMFileSend)
  ON_COMMAND(CM_FILEKERMITRCV,CMFileKermitRcv)
  ON_COMMAND(CM_FILEKERMITGET,CMFileKermitGet)
  ON_COMMAND(CM_FILEKERMITSEND,CMFileKermitSend)
  ON_COMMAND(CM_FILEKERMITFINISH,CMFileKermitFinish)
  ON_COMMAND(CM_FILEXRCV,CMFileXRcv)
  ON_COMMAND(CM_FILEXSEND,CMFileXSend)
  ON_COMMAND(CM_FILEZRCV,CMFileZRcv)
  ON_COMMAND(CM_FILEZSEND,CMFileZSend)
  ON_COMMAND(CM_FILEBPRCV,CMFileBPRcv)
  ON_COMMAND(CM_FILEBPSEND,CMFileBPSend)
  ON_COMMAND(CM_FILEQVRCV,CMFileQVRcv)
  ON_COMMAND(CM_FILEQVSEND,CMFileQVSend)
  ON_COMMAND(CM_FILECHANGEDIR,CMFileChangeDir)
  ON_COMMAND(CM_FILEPRINT,CMFilePrint)
  ON_COMMAND(CM_FILEEXIT,CMFileExit)
  ON_COMMAND(CM_EDITCOPY,CMEditCopy)
  ON_COMMAND(CM_EDITPASTE,CMEditPaste)
  ON_COMMAND(CM_EDITPASTECR,CMEditPasteCR)
  ON_COMMAND(CM_EDITCLEARSCREEN,CMEditClearScreen)
  ON_COMMAND(CM_EDITCLEARBUFFER,CMEditClearBuffer)
  ON_COMMAND(CM_SETUPTERMINAL,CMSetupTerminal)
  ON_COMMAND(CM_SETUPWINDOW,CMSetupWindow)
  ON_COMMAND(CM_SETUPFONT,CMSetupFont)
  ON_COMMAND(CM_SETUPKEYBOARD,CMSetupKeyboard)
  ON_COMMAND(CM_SETUPSERIALPORT,CMSetupSerialPort)
  ON_COMMAND(CM_SETUPTCPIP,CMSetupTCPIP)
  ON_COMMAND(CM_SETUPGENERAL,CMSetupGeneral)
  ON_COMMAND(CM_SETUPSAVE,CMSetupSave)
  ON_COMMAND(CM_SETUPRESTORE,CMSetupRestore)
  ON_COMMAND(CM_CONTROLRESETTERMINAL,CMControlResetTerminal)
  ON_COMMAND(CM_CONTROLAREYOUTHERE,CMControlAreYouThere)
  ON_COMMAND(CM_CONTROLSENDBREAK,CMControlSendBreak)
  ON_COMMAND(CM_CONTROLRESETPORT,CMControlResetPort)
  ON_COMMAND(CM_CONTROLOPENTEK,CMControlOpenTEK)
  ON_COMMAND(CM_CONTROLCLOSETEK,CMControlCloseTEK)
  ON_COMMAND(CM_CONTROLMACRO,CMControlMacro)
  ON_COMMAND(CM_HELPINDEX,CMHelpIndex)
  ON_COMMAND(CM_HELPUSING,CMHelpUsing)
  ON_COMMAND(CM_HELPABOUT,CMHelpAbout)
  //}}AFX_MSG_MAP
END_MESSAGE_MAP()

CVTWindow::CVTWindow()
{
  char Temp[257];
  WNDCLASS wc;
  RECT rect;
  DWORD Style;

  cv.Open = FALSE;
  cv.Ready = FALSE;

  MsgDlgHelp = RegisterWindowMessage(HELPMSGSTRING);

  StartTeraTerm(&ts);

  /* Parse command line parameters*/
  if (LoadTTSET())
    (*ParseParam)(GetCommandLine(), &ts, &(TopicName[0]));
  FreeTTSET();

  InitKeyboard();
  SetKeyMap();

/*  if WinSet.Dlg3d>0 then
    WinSet.Dlg3d = word(LoadCtl3d); */

  AdjustSize = TRUE;

  /* window status */
  Active = FALSE;
  Minimized = FALSE;
  DblClk = FALSE;

  FirstPaint = TRUE;
  /* Initialize scroll buffer */
  InitBuffer();

  InitDisp();

  if (ts.HideTitle>0)
    Style = WS_POPUP | WS_THICKFRAME |
            WS_BORDER | WS_VSCROLL | WS_HSCROLL;
  else
    Style = WS_VSCROLL | WS_HSCROLL |
            WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX |
            WS_BORDER | WS_THICKFRAME;

  wc.style = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = AfxWndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = AfxGetInstanceHandle();
  wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IdVTIcon));
  wc.hCursor = LoadCursor(NULL,IDC_IBEAM);
  wc.hbrBackground = NULL;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = "VTWin32";

  RegisterClass(&wc);
  LoadAccelTable(MAKEINTRESOURCE(Id_Accs));

  if (ts.VTPos.x==CW_USEDEFAULT)
    rect = rectDefault;
  else {
    rect.left = ts.VTPos.x;
    rect.top = ts.VTPos.y;
    rect.right = rect.left + 100;
    rect.bottom = rect.top + 100;
  }
  Create("VTWin32", "Tera Term", Style, rect, NULL, NULL);

  /*--------- Init2 -----------------*/
  HVTWin = GetSafeHwnd();
  if (HVTWin == NULL) return;

  if ((ts.HideTitle==0) && (ts.PopupMenu==0))
  {
    InitMenu(&MainMenu);
    ::SetMenu(HVTWin,MainMenu);
  }
  else
    MainMenu = 0;

  if (! ts.Ok && LoadTTDLG())
  {
    if ((*SetupGeneral)(HVTWin,&ts.PortType,&ts.Language))
    {
      FreeTTDLG();

      /* for Windows NT J */
      if ((ts.Language==IdJapanese) &&
	  (stricmp(ts.VTFont,"Terminal")==0))
	ts.VTFontCharSet = 128;

      if (LoadTTSET())
      {
        (*WriteIniFile)(ts.SetupFName,&ts);
        FreeTTSET();
        ts.Ok = TRUE;
        ChangeDefaultSet(&ts);
      }
    }
    else
      FreeTTDLG();
    if (! ts.Ok) PostQuitMessage(0);
  }

  /* Reset Terminal */
  ResetTerminal();

  if (ts.PopupMenu>0)
    ::PostMessage(HVTWin,WM_USER+MsgChangeMenu,0,0);

  ChangeFont();

  ResetIME();

  ChangeWindowSize(NumOfColumns,NumOfLines);

  WinWidthOld = NumOfColumns;
  WinHeightOld = NumOfLines;

  ChangeTitle();
  ::ShowWindow(HVTWin, SW_SHOWNORMAL);
  /* Enable drag-drop */
  ::DragAcceptFiles(HVTWin,TRUE);
}

LRESULT CVTWindow::DefWindowProc(UINT message, WPARAM wParam, LPARAM lParam)
{
  LRESULT Result;

  if (message == MsgDlgHelp)
  {
    OnDlgHelp(wParam,lParam);
    return 0;
  }
  else if ((ts.HideTitle>0) &&
           (message == WM_NCHITTEST))
  {
    Result = CWnd::DefWindowProc(message,wParam,lParam);
    if ((Result==HTCLIENT) && AltKey())
      Result = HTCAPTION;
    return Result;
  }

  return (CWnd::DefWindowProc(message,wParam,lParam));
}

void CVTWindow::InitMenu(HMENU *Menu)
{
  *Menu = LoadMenu(AfxGetInstanceHandle(),
    MAKEINTRESOURCE(Id_Menu));
  FileMenu = GetSubMenu(*Menu,CM_FILE);
  TransMenu = GetSubMenu(FileMenu,CM_TRANSFER);
  EditMenu = GetSubMenu(*Menu,CM_EDIT);
  SetupMenu = GetSubMenu(*Menu,CM_SETUP);
  ControlMenu = GetSubMenu(*Menu,CM_CONTROL);
}

void CVTWindow::OnActivate
  (UINT nState, CWnd* pWndOther, BOOL bMinimized)
{
  if (nState!=WA_INACTIVE)
  {
    Active = TRUE;
    SetFocus();
    if ((ts.Language==IdJapanese) &&
        CanUseIME())
      EnableIME(TRUE);
    ChangeCaret();
    ActiveWin = IdVT;
  }
  else {
    if ((ts.Language==IdJapanese) &&
        CanUseIME())
      /* position & font of conv. window -> default */
      SetConversionWindow(HVTWin,-1,0);
    DestroyCaret();
    Active = FALSE;
  }
}

void CVTWindow::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  int i, Count;
  char Code;

  if (TalkStatus!=IdTalkKeyb) return;
  Code = nChar;

  for (i=1 ; i<=nRepCnt ; i++)
  {
    CommTextOut(&cv,&Code,1);
    if (ts.LocalEcho>0)
      CommTextEcho(&cv,&Code,1);
  }
}

void CVTWindow::OnClose()
{
  if ((HTEKWin!=NULL) && ! ::IsWindowEnabled(HTEKWin))
  {
    MessageBeep(0);
    return;
  }
  if (cv.Ready && (cv.PortType==IdTCPIP) &&
      (ts.ConfirmDisconn!=0) &&
      (::MessageBox(HVTWin,"Disconnect?","Tera Term",MB_YESNO)==IDNO))
    return;

  DestroyWindow();
}

void CVTWindow::OnDestroy()
{
  EndKeyboard();

//  FreeCtl3d();

  /* Disable drag-drop */
  ::DragAcceptFiles(HVTWin,FALSE);

  EndDDE();

  if (cv.TelFlag) EndTelnet();
  CommClose(&cv);

  OpenHelp(HVTWin,HELP_QUIT,0);

  FreeWINNLS();
  FreeTTSET();
  do { }
    while (FreeTTDLG());

  LogEnd();
  FileSendEnd();
  ProtoEnd();
  do {} while (FreeTTFILE());

  if (HTEKWin != NULL)
    ::DestroyWindow(HTEKWin);

  EndDisp();

  FreeBuffer();

  CFrameWnd::OnDestroy();
}

void CVTWindow::OnDropFiles(HDROP hDropInfo)
{
  if ((SendVar==NULL) && NewFileVar(&SendVar))
  {
    if (DragQueryFile(hDropInfo,0,SendVar->FullName,
                      sizeof(SendVar->FullName))>0)
    {
      SendVar->DirLen = 0;
      ts.TransBin = 0;
      FileSendStart();
    }
    else
      FreeFileVar(&SendVar);
  }
  DragFinish(hDropInfo);
}

void CVTWindow::OnGetMinMaxInfo(MINMAXINFO FAR* lpMMI)
{
  lpMMI->ptMaxSize.x = 10000;
  lpMMI->ptMaxSize.y = 10000;
  lpMMI->ptMaxTrackSize.x = 10000;
  lpMMI->ptMaxTrackSize.y = 10000;
}

void CVTWindow::InitMenuPopup(HMENU SubMenu)
{
  if ( SubMenu == FileMenu )
  {
    if ( Connecting )
      EnableMenuItem(FileMenu,CM_FILENEWCONNECTION,MF_BYCOMMAND || MF_GRAYED);
    else
      EnableMenuItem(FileMenu,CM_FILENEWCONNECTION,MF_BYCOMMAND || MF_ENABLED);

    if ( (! cv.Ready) || (LogVar!=NULL) || (cv.PortType==IdFile) )
      EnableMenuItem(FileMenu,CM_FILELOG,MF_BYCOMMAND || MF_GRAYED);
    else
      EnableMenuItem(FileMenu,CM_FILELOG,MF_BYCOMMAND || MF_ENABLED);

    if ( (! cv.Ready) || (SendVar!=NULL) || (FileVar!=NULL) ||
       (cv.PortType==IdFile) )
    {
      EnableMenuItem(FileMenu,CM_FILESEND,MF_BYCOMMAND || MF_GRAYED);
      EnableMenuItem(FileMenu,4,MF_BYPOSITION || MF_GRAYED); /* Transfer */
      EnableMenuItem(FileMenu,CM_FILECHANGEDIR,MF_BYCOMMAND || MF_GRAYED);
      EnableMenuItem(FileMenu,CM_FILEDISCONNECT,MF_BYCOMMAND || MF_GRAYED);
    }
    else {
      EnableMenuItem(FileMenu,CM_FILESEND,MF_BYCOMMAND || MF_ENABLED);
      EnableMenuItem(FileMenu,4,MF_BYPOSITION || MF_ENABLED); /* Transfer */
      EnableMenuItem(FileMenu,CM_FILECHANGEDIR,MF_BYCOMMAND || MF_ENABLED);
      EnableMenuItem(FileMenu,CM_FILEDISCONNECT,MF_BYCOMMAND || MF_ENABLED);
    }
  }
  else if ( SubMenu == TransMenu )
  {
    if ( (cv.PortType==IdSerial) &&
       ((ts.DataBit==IdDataBit7) || (ts.Flow==IdFlowX)) )
    {
      EnableMenuItem(TransMenu,1,MF_BYPOSITION || MF_GRAYED);  /* XMODEM */
      EnableMenuItem(TransMenu,4,MF_BYPOSITION || MF_GRAYED);  /* Quick-VAN */
    }
    else {
      EnableMenuItem(TransMenu,1,MF_BYPOSITION || MF_ENABLED); /* XMODEM */
      EnableMenuItem(TransMenu,4,MF_BYPOSITION || MF_ENABLED); /* Quick-VAN */
    }
    if ( (cv.PortType==IdSerial) &&
       (ts.DataBit==IdDataBit7) )
    {
      EnableMenuItem(TransMenu,2,MF_BYPOSITION || MF_GRAYED); /* ZMODEM */
      EnableMenuItem(TransMenu,3,MF_BYPOSITION || MF_GRAYED); /* B-Plus */
    }
    else {
      EnableMenuItem(TransMenu,2,MF_BYPOSITION || MF_ENABLED); /* ZMODEM */
      EnableMenuItem(TransMenu,3,MF_BYPOSITION || MF_ENABLED); /* B-Plus */
    }
  }
  else if ( SubMenu == EditMenu )
  {
    if ( Select )
      EnableMenuItem(EditMenu,CM_EDITCOPY,MF_BYCOMMAND || MF_ENABLED);
    else
      EnableMenuItem(EditMenu,CM_EDITCOPY,MF_BYCOMMAND || MF_GRAYED);

    if ( cv.Ready &&
         (SendVar==NULL) && (FileVar==NULL) &&
         (cv.PortType!=IdFile) &&
         (IsClipboardFormatAvailable(CF_TEXT) ||
          IsClipboardFormatAvailable(CF_OEMTEXT)) )
    {
      EnableMenuItem(EditMenu,CM_EDITPASTE,MF_BYCOMMAND || MF_ENABLED);
      EnableMenuItem(EditMenu,CM_EDITPASTECR,MF_BYCOMMAND || MF_ENABLED);
    }
    else {
      EnableMenuItem(EditMenu,CM_EDITPASTE,MF_BYCOMMAND || MF_GRAYED);
      EnableMenuItem(EditMenu,CM_EDITPASTECR,MF_BYCOMMAND || MF_GRAYED);
    }
  }
  else if ( SubMenu == SetupMenu )
    if ( cv.Ready &&
       ((cv.PortType==IdTCPIP) || (cv.PortType==IdFile)) ||
        (SendVar!=NULL) || (FileVar!=NULL) || Connecting )
      EnableMenuItem(SetupMenu,CM_SETUPSERIALPORT,MF_BYCOMMAND || MF_GRAYED);
    else
      EnableMenuItem(SetupMenu,CM_SETUPSERIALPORT,MF_BYCOMMAND || MF_ENABLED);

  else if ( SubMenu == ControlMenu )
  {
    if ( cv.Ready &&
         (SendVar==NULL) && (FileVar==NULL) )
    {
      EnableMenuItem(ControlMenu,CM_CONTROLSENDBREAK,MF_BYCOMMAND || MF_ENABLED);
      if (cv.PortType==IdSerial)
        EnableMenuItem(ControlMenu,CM_CONTROLRESETPORT,MF_BYCOMMAND || MF_ENABLED);
    }
    else {
      EnableMenuItem(ControlMenu,CM_CONTROLSENDBREAK,MF_BYCOMMAND || MF_GRAYED);
      EnableMenuItem(ControlMenu,CM_CONTROLRESETPORT,MF_BYCOMMAND || MF_GRAYED);
    }

    if ( cv.Ready && cv.TelFlag && (FileVar==NULL) )
      EnableMenuItem(ControlMenu,CM_CONTROLAREYOUTHERE,MF_BYCOMMAND || MF_ENABLED);
    else
      EnableMenuItem(ControlMenu,CM_CONTROLAREYOUTHERE,MF_BYCOMMAND || MF_GRAYED);

    if ( HTEKWin==0 ) EnableMenuItem(ControlMenu,CM_CONTROLCLOSETEK,MF_BYCOMMAND || MF_GRAYED);
                 else EnableMenuItem(ControlMenu,CM_CONTROLCLOSETEK,MF_BYCOMMAND || MF_ENABLED);

    if ( (! cv.Ready) || (ConvH!=0) || (FileVar!=NULL) )
      EnableMenuItem(ControlMenu,CM_CONTROLMACRO,MF_BYCOMMAND || MF_GRAYED);
    else
      EnableMenuItem(ControlMenu,CM_CONTROLMACRO,MF_BYCOMMAND || MF_ENABLED);

  }

}

void CVTWindow::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
  switch (nSBCode) {
    case SB_BOTTOM:
      NewOrgX = NumOfColumns-WinWidth;
      break;
    case SB_ENDSCROLL: return;
    case SB_LINEDOWN: NewOrgX = WinOrgX + 1; break;
    case SB_LINEUP: NewOrgX = WinOrgX - 1; break;
    case SB_PAGEDOWN:
      NewOrgX = WinOrgX + WinWidth - 1;
      break;
    case SB_PAGEUP:
      NewOrgX = WinOrgX - WinWidth + 1;
      break;
    case SB_THUMBPOSITION: NewOrgX = nPos; break;
    case SB_THUMBTRACK: NewOrgX = nPos; break;
    case SB_TOP: NewOrgX = 0; break;
    default:
      return;
  }
  UpdateScroll();
}

void CVTWindow::OnInitMenuPopup
  (CMenu* pPopupMenu,UINT nIndex, BOOL bSysmenu)
{
  InitMenuPopup(pPopupMenu->m_hMenu);
}

void CVTWindow::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  KeyDown(HVTWin,nChar,nRepCnt,nFlags & 0x1ff);
}

void CVTWindow::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  KeyUp(nChar);
}

void CVTWindow::OnLButtonDblClk(UINT nFlags, CPoint point)
{
  int X, Y, IStart, IEnd;
  LONG TmpPtr;

  DblClkX = point.x;
  DblClkY = point.y;
  X = point.x / FontWidth + WinOrgX;
  Y = point.y / FontHeight + WinOrgY + PageStart;
  if (X<0) X = 0;
  if (X>=NumOfColumns) X = NumOfColumns-1;

  if ((Y>=0) && (Y<BuffEnd))
  {
    LockBuffer();
    TmpPtr = GetLinePtr(Y);
    IStart = X;
    while ((IStart>0) && (CodeBuff[TmpPtr+IStart]!=0x20))
      IStart--;
    if (CodeBuff[TmpPtr+IStart]==0x20) IStart++;
    IEnd = X;
    while ((IEnd<NumOfColumns) && (CodeBuff[TmpPtr+IEnd]!=0x20))
      IEnd++;
    if (IStart<=X)
    {
      SelectStart.x = IStart;
      SelectStart.y = Y;
      SelectEnd.x = IEnd;
      SelectEnd.y = Y;
      SelectEndOld = SelectStart;
      Select = TRUE;
      ChangeSelectRegion();
      /* Copy to the clipboard */
      CBCopy(HVTWin);
    }
    UnlockBuffer();
  }
  DblClk = TRUE;
  ::SetTimer(HVTWin, IdDblClkTimer, GetDoubleClickTime(), NULL);
}

void CVTWindow::ButtonDown(POINT p, int LMR)
{
  HMENU PopupMenu, PopupBase;
  int X, Y;

  if ((LMR==IdLeftButton) && ControlKey() && (MainMenu==NULL))
  {
    ::ClientToScreen(HVTWin, &p);
    InitMenu(&PopupMenu);
    InitMenuPopup(FileMenu);
    InitMenuPopup(EditMenu);
    InitMenuPopup(SetupMenu);
    InitMenuPopup(ControlMenu);
    PopupBase = CreatePopupMenu();
    AppendMenu(PopupBase, MF_STRING | MF_ENABLED | MF_POPUP,
               (UINT)FileMenu, "&File");
    AppendMenu(PopupBase, MF_STRING | MF_ENABLED | MF_POPUP,
               (UINT)EditMenu, "&Edit");
    AppendMenu(PopupBase, MF_STRING | MF_ENABLED | MF_POPUP,
               (UINT)SetupMenu, "&Setup");
    AppendMenu(PopupBase, MF_STRING | MF_ENABLED | MF_POPUP,
               (UINT)ControlMenu, "C&ontrol");
    AppendMenu(PopupBase, MF_STRING | MF_ENABLED | MF_POPUP,
               (UINT)GetSubMenu(PopupMenu,4), "&Help");
    TrackPopupMenu(PopupBase,TPM_LEFTALIGN | TPM_LEFTBUTTON,
                   p.x,p.y,0,HVTWin,NULL);
    DestroyMenu(PopupBase);
    DestroyMenu(PopupMenu);
    PopupMenu = 0;
    return;
  }

  X = (p.x + FontWidth / 2)
       / FontWidth + WinOrgX;
  Y = p.y / FontHeight + WinOrgY + PageStart;
  if (X<0) X = 0;
  if (X>=NumOfColumns) X = NumOfColumns-1;

  if ((Y>=0) && (Y<BuffEnd))
  {
    if (DblClk && (LMR==IdLeftButton) &&
       (abs(p.x-DblClkX)<=GetSystemMetrics(SM_CXDOUBLECLK)) &&
       (abs(p.y-DblClkY)<=GetSystemMetrics(SM_CYDOUBLECLK)))
    {  /* triple click */
      ::KillTimer(HVTWin, IdDblClkTimer);
      DblClk = FALSE;
      LockBuffer();
      SelectEnd = SelectStart;
      ChangeSelectRegion();
      SelectStart.x = 0;
      SelectStart.y = Y;
      SelectEnd.x = NumOfColumns;
      SelectEnd.y = Y;
      SelectEndOld = SelectStart;
      Select = TRUE;
      ChangeSelectRegion();
      CBCopy(HVTWin);
      UnlockBuffer();
      TplClk = TRUE;
    }
    else {
      if (! (LButton || MButton || RButton))
      {
        SelectEndOld = SelectEnd;
        SelectEnd = SelectStart;

        LockBuffer();
        ChangeSelectRegion();
        UnlockBuffer();

        SelectStart.x = X;
        SelectStart.y = Y;
        SelectEnd = SelectStart;
        SelectEndOld = SelectEnd;
        CaretOff();
        TplClk = FALSE;
        Select = TRUE;
        /* for AutoScrolling */
        ::SetCapture(HVTWin);
        ::SetTimer(HVTWin, IdScrollTimer, 100, NULL);
      }
      switch (LMR) {
        case IdRightButton: RButton = TRUE; break;
        case IdMiddleButton: MButton = TRUE; break;
        case IdLeftButton: LButton = TRUE; break;
      }
    }
  }
}

void CVTWindow::OnLButtonDown(UINT nFlags, CPoint point)
{
  POINT p;

  p.x = point.x;
  p.y = point.y;
  ButtonDown(p,IdLeftButton);
}

void ButtonUp(BOOL Paste)
{
  /* disable autoscrolling */
  ::KillTimer(HVTWin,IdScrollTimer);
  ReleaseCapture();

  Select = (SelectStart.x!=SelectEnd.x) ||
	   (SelectStart.y!=SelectEnd.y);
  LButton = FALSE;
  MButton = FALSE;
  RButton = FALSE;
  CaretOn();

  if (Select)
  {
    if ((SelectEnd.y < SelectStart.y) ||
        (SelectEnd.y == SelectStart.y) &&
        (SelectEnd.x < SelectStart.x))
    {
      SelectEndOld = SelectStart;
      SelectStart = SelectEnd;
      SelectEnd = SelectEndOld;
    }
    /* copy to the clipboard */
    LockBuffer();
    CBCopy(HVTWin);
    UnlockBuffer();
  }
  if (Paste)
    CBStartPaste(HVTWin,FALSE);
}

void CVTWindow::OnLButtonUp(UINT nFlags, CPoint point)
{
  if (! LButton) return;
  ButtonUp(FALSE);
}

void CVTWindow::OnMButtonDown(UINT nFlags, CPoint point)
{
  POINT p;

  p.x = point.x;
  p.y = point.y;
  ButtonDown(p,IdMiddleButton);
}

void CVTWindow::OnMButtonUp(UINT nFlags, CPoint point)
{
  if (! MButton) return;
  ButtonUp(TRUE);
}

void CVTWindow::OnMouseMove(UINT nFlags, CPoint point)
{
  int X, Y;

  if (! (LButton | MButton | RButton)) return;

  X = (point.x + FontWidth / 2)
       / FontWidth + WinOrgX;
  Y = point.y / FontHeight + WinOrgY + PageStart;

  SelectEnd.x = X;
  SelectEnd.y = Y;

  if ((SelectEnd.y < SelectStart.y) ||
      (SelectEnd.y==SelectStart.y) &&
      (SelectEnd.x<SelectStart.x))
  {
    SelectEnd.x++;
    if (SelectEnd.x>=NumOfColumns)
    {
      SelectEnd.x = 0;
      SelectEnd.y++;
    }
  }

  if (SelectEnd.x < 0) SelectEnd.x = 0;
  if (SelectEnd.x > NumOfColumns)
    SelectEnd.x = NumOfColumns;

  if (SelectEnd.y < 0) SelectEnd.y = 0;
  if (SelectEnd.y >= BuffEnd)
    SelectEnd.y = BuffEnd - 1;

  LockBuffer();
  ChangeSelectRegion();
  UnlockBuffer();
}

void CVTWindow::OnMove(int x, int y)
{
  int CaretX, CaretY;
  POINT Point;
  RECT R;

  ::GetWindowRect(HVTWin,&R);
  ts.VTPos.x = R.left;
  ts.VTPos.y = R.top;

  if (CanUseIME() && (ts.IMEInline > 0))
  {
    CaretX = (CursorX-WinOrgX)*FontWidth;
    CaretY = (CursorY-WinOrgY)*FontHeight;
    /* set IME conversion window pos. */
    SetConversionWindow(HVTWin,CaretX,CaretY);
  }

  Point.x = 0;
  Point.y = ScreenHeight;
  ::ClientToScreen(HVTWin,&Point);
  CompletelyVisible = (Point.y <= CRTHeight);
}

void CVTWindow::OnNcLButtonDblClk(UINT nHitTest, CPoint point)
{
  if (! Minimized && (nHitTest == HTCAPTION))
  {
    if ((WinWidth==NumOfColumns) && (WinHeight==NumOfLines))
    {
      if (WinWidthOld > NumOfColumns)
        WinWidthOld = NumOfColumns;
      if (WinHeightOld > BuffEnd)
        WinHeightOld = BuffEnd;
      if (ts.TermIsWin==0)
        ChangeWindowSize(WinWidthOld,WinHeightOld);
    }
    else {
      WinWidthOld = WinWidth;
      WinHeightOld = WinHeight;
      ChangeWindowSize(NumOfColumns,NumOfLines);
    }
  }
  else
    CWnd::OnNcLButtonDblClk(nHitTest,point);
}

void CVTWindow::OnNcRButtonDown(UINT nHitTest, CPoint point)
{
  if ((nHitTest==HTCAPTION) &&
      (ts.HideTitle>0) &&
      AltKey())
    ::CloseWindow(HVTWin); /* iconize */
}

void CVTWindow::OnPaint()
{
  PAINTSTRUCT ps;
  CDC *cdc;
  HDC PaintDC;

  cdc = BeginPaint(&ps);
  PaintDC = cdc->GetSafeHdc();

  PaintWindow(PaintDC,&ps);

  if (FirstPaint)
  {
    if (strlen(TopicName)>0)
    {
      InitDDE();
      SendDDEReady();
    }
    FirstPaint = FALSE;
    ::PostMessage(HVTWin,WM_USER+MsgComStart,0,0);
  }

  EndPaint(&ps);
}

void CVTWindow::OnRButtonDown(UINT nFlags, CPoint point)
{
  POINT p;

  p.x = point.x;
  p.y = point.y;
  ButtonDown(p,IdRightButton);
}

void CVTWindow::OnRButtonUp(UINT nFlags, CPoint point)
{
  if (! RButton) return;
  ButtonUp(TRUE);
}

void CVTWindow::OnSize(UINT nType, int cx, int cy)
{
  RECT R;
  int w, h;

  Minimized = (nType==SIZE_MINIMIZED);

  if (FirstPaint && Minimized)
  {
    if (strlen(TopicName)>0)
    {
      InitDDE();
      SendDDEReady();
    }
    FirstPaint = FALSE;
    ::PostMessage(HVTWin,WM_USER+MsgComStart,0,0);
    return;
  }
  if (Minimized || DontChangeSize) return;

  ::GetWindowRect(HVTWin,&R);
  w = R.right - R.left;
  h = R.bottom - R.top;
  ResizeWindow(R.left,R.top,w,h,cx,cy);
}

void CVTWindow::OnSysChar(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  char e = ESC;
  char Code;
  int Count, i;

  if (ts.MetaKey>0)
  {
    if (TalkStatus!=IdTalkKeyb) return;
    Code = nChar;
    for (i=1 ; i<=nRepCnt ; i++)
    {
      CommTextOut(&cv,&e,1);
      CommTextOut(&cv,&Code,1);
      if (ts.LocalEcho>0)
      {
        CommTextEcho(&cv,&e,1);
        CommTextEcho(&cv,&Code,1);
      }
    }
    return;
  }
  CWnd::OnSysChar(nChar,nRepCnt,nFlags);
}

void CVTWindow::OnSysCommand(UINT nID, LPARAM lParam)
{
  if (nID==CM_SHOWMENUBAR)
  {
    ts.PopupMenu = 0;
    SwitchMenu();
  }
  else CWnd::OnSysCommand(nID,lParam);
}

void CVTWindow::OnSysKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  if (nChar==VK_F10)
    OnKeyDown(nChar,nRepCnt,nFlags);
  else
    CWnd::OnSysKeyDown(nChar,nRepCnt,nFlags);
}

void CVTWindow::OnSysKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  if (nChar==VK_F10)
    OnKeyUp(nChar,nRepCnt,nFlags);
  else
    CWnd::OnSysKeyUp(nChar,nRepCnt,nFlags);
}

void CVTWindow::OnTimer(UINT nIDEvent)
{
  POINT Point;

  if (nIDEvent==IdScrollTimer)
  {
    GetCursorPos(&Point);
    ::ScreenToClient(HVTWin,&Point);
    AutoScroll(Point);
    if ((Point.x < 0) || (Point.x >= ScreenWidth) ||
        (Point.y < 0) || (Point.y >= ScreenHeight))
      ::PostMessage(HVTWin,WM_MOUSEMOVE,MK_LBUTTON,MAKELONG(Point.x,Point.y));
    return;
  }

  ::KillTimer(HVTWin, nIDEvent);

  switch (nIDEvent) {
    case IdDelayTimer: cv.CanSend = TRUE; break;
    case IdProtoTimer: if (ProtoFlag) ProtoDlgTimeOut();
    case IdDblClkTimer: DblClk = FALSE; break;
    case IdComEndTimer:
      if (cv.TelFlag) EndTelnet();
      CommClose(&cv);
      SetDdeComReady(0);
      if ((ts.PortType==IdTCPIP) &&
          (ts.AutoWinClose>0) &&
          ::IsWindowEnabled(HVTWin) &&
          ((HTEKWin==0) || ::IsWindowEnabled(HTEKWin)) )
        DestroyWindow();
      else
        ChangeTitle();
      break;
  }
}

void CVTWindow::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
  switch (nSBCode) {
    case SB_BOTTOM:
      NewOrgY = BuffEnd-WinHeight-PageStart;
      break;
    case SB_ENDSCROLL: return;
    case SB_LINEDOWN: NewOrgY = WinOrgY + 1; break;
    case SB_LINEUP: NewOrgY = WinOrgY - 1; break;
    case SB_PAGEDOWN:
      NewOrgY = WinOrgY + WinHeight - 1;
      break;
    case SB_PAGEUP:
      NewOrgY = WinOrgY - WinHeight + 1;
      break;
    case SB_THUMBPOSITION: NewOrgY = nPos-PageStart; break;
    case SB_THUMBTRACK: NewOrgY = nPos-PageStart; break;
    case SB_TOP: NewOrgY = -PageStart; break;
    default:
      return;
  }
  UpdateScroll();
}

LONG CVTWindow::OnIMEReport(UINT wParam, LONG lParam)
{
  LPSTR Str;
  int Len;
  MSG TM;

  switch (wParam) {
    case IR_OPENCONVERT:
      OpenConv = TRUE;
      break;
    case IR_CLOSECONVERT:
      ::PostMessage(HVTWin,WM_USER+MsgCloseIME,0,0);
      break;
    case IR_STRING:
      Str = (LPSTR)GlobalLock((HGLOBAL)lParam);
      Len = strlen(Str);
      if (Len==1)
      {
	switch (Str[0]) {
	  case 0x20:
	    if (ControlKey()) Str[0] = 0; /* Ctrl-Space */
	    break;
	}
      }
      if ( ts.LocalEcho>0 )
        CommTextEcho(&cv,Str,Len);
      CommTextOut(&cv,Str,Len);
      GlobalUnlock((HGLOBAL)lParam);
      return 1;
  }
  return 0;
}

LONG CVTWindow::OnAccelCommand(UINT wParam, LONG lParam)
{
  switch (wParam) {
    case IdAccEditCopy:
      CMEditCopy();
      break;
    case IdAccEditPaste:
      CMEditPaste();
      break;
    case IdAccEditPasteCR:
      CMEditPasteCR();
      break;
    case IdAccLineUp:
      OnVScroll(SB_LINEUP,0,NULL);
      break;
    case IdAccLineDown:
      OnVScroll(SB_LINEDOWN,0,NULL);
      break;
    case IdAccPageUp:
      OnVScroll(SB_PAGEUP,0,NULL);
      break;
    case IdAccPageDown:
      OnVScroll(SB_PAGEDOWN,0,NULL);
      break;
    case IdAccBuffTop:
      OnVScroll(SB_TOP,0,NULL);
      break;
    case IdAccBuffBottom:
      OnVScroll(SB_BOTTOM,0,NULL);
      break;
    case IdAccSendBreak:
      CMControlSendBreak();
      break;
  }
  return 0;
}

LONG CVTWindow::OnChangeMenu(UINT wParam, LONG lParam)
{
  HMENU SysMenu;

  if ((ts.PopupMenu>0) && (MainMenu!=NULL) ||
      (ts.PopupMenu==0) && (MainMenu==NULL))
  {
    if (ts.PopupMenu>0)
    {
      DestroyMenu(MainMenu);
      MainMenu = NULL;
    }
    else
      InitMenu(&MainMenu);

    WinWidthOld = WinWidth;
    WinHeightOld = WinHeight;
    AdjustSize = TRUE;
    ::SetMenu(HVTWin, MainMenu);
    ::DrawMenuBar(HVTWin);
  }

  ::GetSystemMenu(HVTWin,TRUE);
  if (ts.PopupMenu>0)
  {
    SysMenu = ::GetSystemMenu(HVTWin,FALSE);
    AppendMenu(SysMenu, MF_SEPARATOR, 0, NULL);
    AppendMenu(SysMenu, MF_STRING, CM_SHOWMENUBAR, "Show menu &bar");
  }
  return 0;
}

LONG CVTWindow::OnCloseIME(UINT wParam, LONG lParam)
{
  OpenConv = FALSE;
  return 0;
}

LONG CVTWindow::OnComNotify(UINT wParam, LONG lParam)
{
  switch (LOWORD(lParam)) {
    case FD_READ:
      CommProcRRQ(&cv);
      break;
    case FD_CLOSE:
      Connecting = FALSE;
      cv.Ready = FALSE;
      SetTimer(IdComEndTimer,1,NULL);
  }
  return 0;
}

LONG CVTWindow::OnComOpen(UINT wParam, LONG lParam)
{
  CommStart(&cv,lParam);
  Connecting = FALSE;
  ChangeTitle();
  if (! cv.Ready) return 0;

  SetDdeComReady(1);
  if (cv.TelFlag)
  {
    InitTelnet();
    /* Telnet Option Negotiation */

    TelEnableMyOpt(TERMTYPE);

    TelEnableHisOpt(SGA);

    TelEnableMyOpt(SGA);

    TelEnableHisOpt(ECHO);

    TelEnableMyOpt(NAWS);
    if (ts.TelBin>0)
    {
      TelEnableMyOpt(BINARY);
      TelEnableHisOpt(BINARY);
    }
  }

  if (DDELog || FileLog)
  {
    if (! CreateLogBuf())
    {
      if (DDELog) EndDDE();
      if (FileLog) LogEnd();
    }
  }

  if (BinLog)
  {
    if (! CreateBinBuf())
      LogEnd();
  }
  return 0;
}

LONG CVTWindow::OnComStart(UINT wParam, LONG lParam)
{
  if ((ts.PortType!=IdSerial) && (ts.HostName[0]==0))
    CMFileNewConnection();
  else {
    Connecting = TRUE;
    ChangeTitle();
    CommOpen(HVTWin,&ts,&cv);
  }
  return 0;
}

LONG CVTWindow::OnDlgHelp(UINT wParam, LONG lParam)
{
  OpenHelp(HVTWin,HELP_CONTEXT,HelpId);
  return 0;
}

LONG CVTWindow::OnDdeEnd(UINT wParam, LONG lParam)
{
  EndDDE();
  if (CloseTT) OnClose();
  return 0;
}

void CVTWindow::CMEditCopy()
{
  LockBuffer();
  CBCopy(HVTWin);
  UnlockBuffer();
}

void CVTWindow::CMFileNewConnection()
{
  char Command[91], Command2[91];
  TGetHNRec GetHNRec; /* record for dialog box */
  int i;
  char FName[MAXPATHLEN], KeyCnfFN[MAXPATHLEN];
  char Temp[7];

  if (Connecting) return;

  HelpId = HlpFileNewConnection;
  GetHNRec.SetupFN = ts.SetupFName;
  GetHNRec.PortType = ts.PortType;
  GetHNRec.Telnet = ts.Telnet;
  GetHNRec.TCPPort = ts.TCPPort;

  strcpy(Command,"ttermpro ");
  GetHNRec.HostName = &Command[9];

  if (! LoadTTDLG()) return;
  if ((*GetHostName)(HVTWin,&GetHNRec))
  {
    if (! cv.Ready)
    {
      ts.PortType = GetHNRec.PortType;
      ts.Telnet = GetHNRec.Telnet;
      ts.TCPPort = GetHNRec.TCPPort;
      ts.ComPort = GetHNRec.ComPort;

      FName[0] = 0;
      if (LoadTTSET())
        (*ParseParam)(Command, &ts, NULL);
      FreeTTSET();
      if (strlen(FName)>0) strcpy(ts.SetupFName,FName);
      SetKeyMap();
      Connecting = TRUE;
      ChangeTitle();
      CommOpen(HVTWin,&ts,&cv);
      ResetSetup();
    }
    else {
      if (GetHNRec.PortType==IdSerial)
      {
        Command[8] = 0;
        strcat(Command," /C=");
        i = strlen(Command);
        Command[i] = (char)(GetHNRec.ComPort+0x30);
        Command[i+1] = 0;
      }
      else {
        strcpy(Command2, &Command[9]);
        Command[9] = 0;
        if (GetHNRec.Telnet==0)
          strcat(Command," /T=0");
        else
          strcat(Command," /T=1");
        if (GetHNRec.TCPPort<65535)
        {
	  sprintf(Temp," /P=%u",GetHNRec.TCPPort);
          strcat(Command,Temp);                    
        }
        strcat(Command," ");
        strncat(Command, Command2, sizeof(Command)-1);
      }
      WinExec(Command,SW_SHOW);
    }
  }
  FreeTTDLG();
}

void CVTWindow::CMFileLog()
{
  HelpId = HlpFileLog;
  LogStart();
}

void CVTWindow::CMFileSend()
{
  HelpId = HlpFileSend;
  FileSendStart();
}

void CVTWindow::CMFileKermitRcv()
{
  KermitStart(IdKmtReceive);
}

void CVTWindow::CMFileKermitGet()
{
  HelpId = HlpFileKmtGet;
  KermitStart(IdKmtGet);
}

void CVTWindow::CMFileKermitSend()
{
  HelpId = HlpFileKmtSend;
  KermitStart(IdKmtSend);
}

void CVTWindow::CMFileKermitFinish()
{
  KermitStart(IdKmtFinish);
}

void CVTWindow::CMFileXRcv()
{
  HelpId = HlpFileXmodemRecv;
  XMODEMStart(IdXReceive);
}

void CVTWindow::CMFileXSend()
{
  HelpId = HlpFileXmodemSend;
  XMODEMStart(IdXSend);
}

void CVTWindow::CMFileZRcv()
{
  ZMODEMStart(IdZReceive);
}

void CVTWindow::CMFileZSend()
{
  HelpId = HlpFileZmodemSend;
  ZMODEMStart(IdZSend);
}

void CVTWindow::CMFileBPRcv()
{
  BPStart(IdBPReceive);
}

void CVTWindow::CMFileBPSend()
{
  HelpId = HlpFileBPlusSend;
  BPStart(IdBPSend);
}

void CVTWindow::CMFileQVRcv()
{
  QVStart(IdQVReceive);
}

void CVTWindow::CMFileQVSend()
{
  HelpId = HlpFileQVSend;
  QVStart(IdQVSend);
}

void CVTWindow::CMFileChangeDir()
{
  HelpId = HlpFileChangeDir;
  if (! LoadTTDLG()) return;
  (*ChangeDirectory)(HVTWin,ts.FileDir);
  FreeTTDLG();
}

void CVTWindow::CMFilePrint()
{
  HelpId = HlpFilePrint;
  VTPrintInit(FALSE);
}

void CVTWindow::CMFileDisconnect()
{
  ::PostMessage(HVTWin, WM_USER+MsgComNotify, 0, FD_CLOSE);
}

void CVTWindow::CMFileExit()
{
  OnClose();
}

void CVTWindow::CMEditPaste()
{
  CBStartPaste(HVTWin,FALSE);
}

void CVTWindow::CMEditPasteCR()
{
  CBStartPaste(HVTWin,TRUE);
}

void CVTWindow::CMEditClearScreen()
{
  LockBuffer();
  BuffScroll(NumOfLines);
  MoveCursor(0,0);
  UnlockBuffer();
  WinOrgY = WinOrgY-NumOfLines;
  ScrollHomePos();
  CaretOn();
}

void CVTWindow::CMEditClearBuffer()
{
  LockBuffer();
  ClearBuffer();
  UnlockBuffer();
  CaretOn();
}

void CVTWindow::CMSetupTerminal()
{
  HelpId = HlpSetupTerminal;
  if (! LoadTTDLG()) return;
  (*SetupTerminal)(HVTWin, &ts,ts.Language);
  FreeTTDLG();
  if (ts.Ok) SetupTerm();
}

void CVTWindow::CMSetupWindow()
{
  HelpId = HlpSetupWindow;
  ts.VTFlag = TRUE;
  ts.SampleFont = VTFont[0];

  if (! LoadTTDLG()) return;
  (*SetupWin)(HVTWin, &ts);
  FreeTTDLG();

  if (ts.Ok) ChangeWin();
}

void CVTWindow::CMSetupFont()
{
  BOOL Ok;

  HelpId = HlpSetupFont;
  if (! LoadTTDLG()) return;
  SetLogFont();
  Ok = (*ChooseFontDlg)(HVTWin,&VTlf,&ts.EnableBold);
  FreeTTDLG();
  if (! Ok) return;

  strcpy(ts.VTFont,VTlf.lfFaceName);
  ts.VTFontSize.x = VTlf.lfWidth;
  ts.VTFontSize.y = VTlf.lfHeight;
  ts.VTFontCharSet = VTlf.lfCharSet;

  ChangeFont();

  ChangeWindowSize(WinWidth,WinHeight);

  ChangeCaret();
}

void CVTWindow::CMSetupKeyboard()
{
  HelpId = HlpSetupKeyboard;
  if (! LoadTTDLG()) return;
  (*SetupKeyboard)(HVTWin, &ts,ts.Language);
  FreeTTDLG();

  if ((ts.Ok) &&
      (ts.Language==IdJapanese))
    ResetIME();
}

void CVTWindow::CMSetupSerialPort()
{
  HelpId = HlpSetupSerialPort;
  if (! LoadTTDLG()) return;
  (*SetupSerialPort)(HVTWin, &ts);
  FreeTTDLG();

  if (ts.Ok)
  {
    if (cv.Open)
    {
      if (ts.ComPort != cv.ComPort)
      {
        CommClose(&cv);
        CommOpen(HVTWin,&ts,&cv);
      }
      else
        CommResetSerial(&ts,&cv);
    }
    else
      CommOpen(HVTWin,&ts,&cv);
  }
}

void CVTWindow::CMSetupTCPIP()
{
  HelpId = HlpSetupTCPIP;
  if (! LoadTTDLG()) return;
  (*SetupTCPIP)(HVTWin, &ts);
  FreeTTDLG();
}

void CVTWindow::CMSetupGeneral()
{
  char Temp[8];

  HelpId = HlpSetupGeneral;
  if (! LoadTTDLG()) return;
  if ((*SetupGeneral)(HVTWin,&ts.PortType,&ts.Language))
  {
    if (ts.PortType == IdSerial)
      strcpy(Temp,"serial");
    else /* IdFile -> IdTCPIP */
      strcpy(Temp,"tcpip");
    WritePrivateProfileString("Tera Term","Port",Temp,ts.SetupFName);
    LoadDefaultSet(ts.SetupFName);

    ResetCharSet();
    ResetIME();
  }
  FreeTTDLG();
}

void CVTWindow::CMSetupSave()
{
  BOOL Ok;
  char TmpSetupFN[MAXPATHLEN];

  strcpy(TmpSetupFN,ts.SetupFName);
  if (! LoadTTFILE()) return;
  HelpId = HlpSetupSave;
  Ok = (*GetSetupFname)(HVTWin,TRUE,&ts);
  FreeTTFILE();
  if (! Ok) return;

  if (LoadTTSET())
  {
    /* write current setup values to file */
    (*WriteIniFile)(ts.SetupFName,&ts);
    /* copy host list */
    (*CopyHostList)(TmpSetupFN,ts.SetupFName);
    FreeTTSET();
  }

  ChangeDefaultSet(&ts);
}

void CVTWindow::ResetSetup()
{
  ChangeFont();
  ChangeWindowSize(WinWidth,WinHeight);
  ChangeCaret();

  /* setup terminal */
  SetupTerm();

  /* setup window */
  ChangeWin();

  /* Language & IME */
  ResetIME();

  /* if serial port, change port parameters */
  if (cv.Ready && (cv.PortType==IdSerial) &&
      (cv.ComPort==ts.ComPort))
    CommResetSerial(&ts,&cv);

  /* change TEK window */
  TEKWinRestoreSetup();
}

void CVTWindow::CMSetupRestore()
{
  int i;
  BOOL Ok;

  HelpId = HlpSetupRestore;
  if (! LoadTTFILE()) return;
  Ok = (*GetSetupFname)(HVTWin,FALSE,&ts);
  FreeTTFILE();
  if (! Ok) return;

  if (LoadTTSET())
    (*ReadIniFile)(ts.SetupFName,&ts);
  FreeTTSET();

  ChangeDefaultSet(&ts);

  ResetSetup();
}

void CVTWindow::CMControlResetTerminal()
{
  ScrollHomePos();
  ResetTerminal();
  CaretOn();
}

void CVTWindow::CMControlAreYouThere()
{
  if (cv.Ready && (cv.PortType==IdTCPIP))
    TelSendAYT();
}

void CVTWindow::CMControlSendBreak()
{
  if (cv.Ready)
    switch (cv.PortType) {
      case IdTCPIP:
	TelSendBreak();
	break;
      case IdSerial:
	CommSendBreak(&cv);
	break;
    }
}

void CVTWindow::CMControlResetPort()
{
  CommResetSerial(&ts,&cv);
}

void CVTWindow::CMControlOpenTEK()
{
  TEKWinActivate();
}

void CVTWindow::CMControlCloseTEK()
{
  if ((HTEKWin==NULL) ||
      ! ::IsWindowEnabled(HTEKWin))
    MessageBeep(0);
  else
    ::DestroyWindow(HTEKWin);
}

void CVTWindow::CMControlMacro()
{
  char Cmnd[40];

  SetTopic();
  if (! InitDDE()) return;
  strcpy(Cmnd,"TTMACRO /D=");
  strcat(Cmnd,TopicName);
  if (WinExec(Cmnd,SW_MINIMIZE) < 32)
    EndDDE();
}

void CVTWindow::CMHelpIndex()
{
  OpenHelp(HVTWin,HELP_INDEX,0);
}

void CVTWindow::CMHelpUsing()
{
  ::WinHelp(HVTWin,"", HELP_HELPONHELP, 0);
}

void CVTWindow::CMHelpAbout()
{
  if (! LoadTTDLG()) return;
  (*AboutDialog)(HVTWin);
  FreeTTDLG();
}
